/*
 *	Copyright 2011 Paul Parsons
 *
 *	This file is part of TinyInit.
 *
 *	TinyInit is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	TinyInit is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with TinyInit.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <getopt.h>
#include <signal.h>
#include <stdio.h>
#if	defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
#include <stdlib.h>		/* Get daemon(3) declaraton */
#endif
#include <string.h>
#include <unistd.h>

#include <sys/sem.h>
#include <sys/shm.h>

#include "init.h"

static enum { MODENULL=-1, START=0, STOP=1, LIST } mode = MODENULL;

static int write_label(const char *label, struct tinyshm *tiny)
{
	int slot;
	int i;

	slot = -1;

	for (i = 0; i < MAX_DMON; ++i)
	{
		if (tiny->dpid[i] == 0)
		{
			if (slot == -1)
				slot = i;
		}
		else if (strncmp(tiny->dn[i].name, label, DMONNAMELEN) == 0)
		{
			slot = i;
			break;
		}
	}

	if (slot == -1)
	{
		errno = ENOSPC;
	}
	else
	{
		tiny->dpid[slot] = getpid();
		strncpy(tiny->dn[slot].name, label, DMONNAMELEN);
	}

	return (slot);
}

static pid_t read_label(const char *label, struct tinyshm *tiny)
{
	int i;
	pid_t pid;

	for (i = 0; i < MAX_DMON; ++i)
	{
		if (tiny->dpid[i] != 0 && strncmp(tiny->dn[i].name, label, DMONNAMELEN) == 0)
			break;
	}

	if (i == MAX_DMON)
	{
		pid = -1;
		errno = ESRCH;
	}
	else
	{
		pid = tiny->dpid[i];
		tiny->dpid[i] = 0;
	}

	return (pid);
}

static int (*const wr_label[2])(const char *label, struct tinyshm *tiny) =
{
	[START]=write_label,
	[STOP]=	read_label,
};

static int shm_access(
	const char *label,
	int (*func)(const char *, struct tinyshm *))
{
	int ret;
	int shmid;
	struct tinyshm *tiny;
	int semid;

	ret = -1;

	shmid = shmget(INIT_KEY, sizeof(struct tinyshm), 0);
	if (shmid == -1)
	{
		perror("shmget");
		goto retdt;
	}
	tiny = shmat(shmid, NULL, 0);
	if (tiny == (struct tinyshm *)-1)
	{
		perror("shmat");
		goto retdt;
	}

	semid = semget(INIT_KEY, 1, 0);
	if (semid == -1)
	{
		perror("semget");
		goto retat;
	}
	if (sem_dn(semid) == -1)
	{
		perror("semop");
		goto retat;
	}

	ret = (*func)(label, tiny);

	(void) sem_up(semid);

retat:
	(void) shmdt(tiny);
retdt:
	return (ret);
}

static int write_file(const char *file)
{
	FILE *fp;

	fp = fopen(file, "w");
	if (! fp)
	{
		perror(file);
		return (-1);
	}

	if (fprintf(fp, "%d\n", getpid()) < 0)
	{
		(void) fclose(fp);
		return (-1);
	}

	(void) fclose(fp);

	return (0);
}

static pid_t read_file(const char *file)
{
	FILE *fp;
	pid_t pid;

	fp = fopen(file, "r");
	if (! fp)
	{
		perror(file);
		return (-1);
	}

	if (fscanf(fp, "%d", &pid) != 1)
	{
		(void) fclose(fp);
		return (-1);
	}

	(void) fclose(fp);

	(void) unlink(file);

	return (pid);
}

static int (*const wr_file[2])(const char *file) =
{
	[START]=write_file,
	[STOP]=	read_file,
};

static int list_label(const char *label, struct tinyshm *tiny)
{
	int i;

	printf("PID\tLABEL\n");
	for (i = 0; i < MAX_DMON; ++i)
	{
		if (tiny->dpid[i] != 0)
		{
			if (kill(tiny->dpid[i], 0) == -1)
				tiny->dpid[i] = 0;
			else
				printf("%d\t%.32s\n", tiny->dpid[i], tiny->dn[i].name);
		}
	}

	return (0);
}

static const struct option long_options[] =
{
	{	"start",	0,	0,	'S'	},
	{	"stop",		0,	0,	'K'	},
	{	"label",	1,	0,	'l'	},
	{	"pidfile",	1,	0,	'p'	},
	{	"signal",	1,	0,	's'	},
	{	"list",		0,	0,	'L'	},
	{	NULL,		0,	NULL,	0	}
};

extern char **environ;

int main(int argc, char *argv[])
{
	char *label;
	char *file;
	int sig;;
	int c;
	int ret;

	label = NULL;
	file = NULL;
	sig = SIGTERM;

	opterr = 0;
	while ((c = getopt_long(argc, argv, "+", long_options, NULL)) != -1)
	{
		switch (c)
		{
		case 'S':
			if (mode != MODENULL)
				goto optdone;
			mode = START;
			break;
		case 'K':
			if (mode != MODENULL)
				goto optdone;
			mode = STOP;
			break;
		case 'l':
			if (mode != START && mode != STOP)
				goto optdone;
			label = optarg;
			if (strlen(label) == 0)	/* i.e. "--label=" */
				goto optdone;
			break;
		case 'p':
			if (mode != START && mode != STOP)
				goto optdone;
			file = optarg;
			if (strlen(file) == 0)	/* i.e. "--pidfile=" */
				goto optdone;
			break;
		case 's':
			if (mode != STOP)
				goto optdone;
			if (strcmp(optarg, "HUP") == 0)
				sig = SIGHUP;
			else if (strcmp(optarg, "TERM") == 0)
				sig = SIGTERM;
			else
				goto optdone;
			break;
		case 'L':
			if (mode != MODENULL)
				goto optdone;
			mode = LIST;
			break;
		}
		if (c == '?')
			break;
	}
optdone:

	if (c != -1
		|| mode == MODENULL
		|| (mode == START && optind == argc)
		|| ((mode == STOP || mode == LIST) && optind != argc)
		|| ((mode == START || mode == STOP) && (! label && ! file))
		|| (label && file))
	{
		fprintf(stderr, "usage:\tdaemon --start --label=label "
			"pathname [arguments]\n");
		fprintf(stderr, "\tdaemon --start --pidfile=file "
			"pathname [arguments]\n");
		fprintf(stderr, "\tdaemon --stop --label=label [--signal=sig]\n");
		fprintf(stderr, "\tdaemon --stop --pidfile=file [--signal=sig]\n");
		fprintf(stderr, "\tdaemon --list\n");
		return (1);
	}

	if (label && strlen(label) > (DMONNAMELEN-1))
	{
		fprintf(stderr, "Label exceeds %d characters\n", (DMONNAMELEN-1));
		return (1);
	}

	if (mode == START)
	{
		/* Cursory check for executable before fork() precludes error */
		if (access(argv[optind], X_OK) == -1)
		{
			perror("access");
			return (1);
		}

		if (daemon(0, 0) == -1)
		{
			perror("daemon");
			return (1);
		}
	}

	if (label)
		ret = shm_access(label, wr_label[mode]);
	else if (file)
		ret = (*wr_file[mode])(file);
	else	/* (mode == LIST) */
		ret = shm_access(NULL, list_label);
	if (ret == -1)
		return (1);

	if (mode == START)
	{
		(void) execve(argv[optind], &argv[optind], environ);
		return (1);
	}

	if (mode == STOP)
	{
		if (kill_kill(ret, sig) == -1)
		{
			fprintf(stderr, "Process %d cannot be killed\n", ret);
			return (1);
		}
		return (0);
	}

	/* (mode == LIST) */
	return (0);
}
