/*
 *	Copyright 2011 Paul Parsons
 *
 *	This file is part of TinyInit.
 *
 *	TinyInit is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	TinyInit is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with TinyInit.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <getopt.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <sys/sem.h>
#include <sys/shm.h>

#include "init.h"

static const struct option long_options[] =
{
	{	"noinit",	0,	0,	'i'	},
	{	"nosync",	0,	0,	's'	},
	{	NULL,		0,	NULL,	0	}
};

static int acknaksig;
static void handler(int sig)
{
	acknaksig = sig;
}

int main(int argc, char *argv[])
{
	char *s;
	enum halttype type;
	int noinit;
	int nosync;
	int c;
	sigset_t mask;
	struct sigaction sa;
	int shmid;
	struct tinyshm *tiny;
	int semid;

	/* Process arguments */

	s = strrchr(argv[0], '/');
	if (! s)
		s = argv[0];
	else
		++s;
	if (strcmp(s, "halt") == 0)
		type = HT_HALT;
	else if (strcmp(s, "poweroff") == 0)
		type = HT_POWEROFF;
	else if (strcmp(s, "reboot") == 0)
		type = HT_REBOOT;
	else
		type = HT_NULL;

	noinit = 0;
	nosync = 0;

	opterr = 0;
	while ((c = getopt_long(argc, argv, "+", long_options, NULL)) != -1)
	{
		switch (c)
		{
		case 'i':
			noinit = 1;
			break;
		case 's':
			nosync = 1;
			break;
		}
		if (c == '?')
			break;
	}

	if (type == HT_NULL || c != -1 || optind != argc)
	{
		fprintf(stderr, "usage:\thalt [--noinit [--nosync]]\n");
		fprintf(stderr, "\tpoweroff [--noinit [--nosync]]\n");
		fprintf(stderr, "\treboot [--noinit [--nosync]]\n");
		return (1);
	}

	if (noinit == 1)
	{
		myreboot(type, nosync);
		perror("halt");
		return (1);
	}

	/* Block ACK/NAK signals */

	sigemptyset(&mask);
	sigaddset(&mask, SIGACK);
	sigaddset(&mask, SIGNAK);
	if (sigprocmask(SIG_SETMASK, &mask, &mask) == -1)
		return (2);

	/* Catch ACK/NAK signals */

	sa.sa_handler = handler;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	if (sigaction(SIGACK, &sa, NULL) == -1)
		return (2);
	if (sigaction(SIGNAK, &sa, NULL) == -1)
		return (2);

	/* Get shm */

	shmid = shmget(INIT_KEY, sizeof(struct tinyshm), 0);
	if (shmid == -1)
	{
		perror("shmget");
		return (3);
	}
	tiny = shmat(shmid, NULL, 0);
	if (tiny == (struct tinyshm *)-1)
	{
		perror("shmat");
		return (3);
	}

	/* Get sem */

	semid = semget(INIT_KEY, 1, 0);
	if (semid == -1)
	{
		perror("semget");
		return (4);
	}
	if (sem_dn(semid) == -1)
	{
		perror("semop");
		return (4);
	}

	/* Write shm */

	tiny->ccmd = CC_HALT;
	tiny->cpid = getpid();
	tiny->u.type = type;
	tiny->cseq += 1;	/* Must be last field written */
	(void) shmdt(tiny);

	/* Poke init */

	if (kill(1, SIGPOKE) == -1)
	{
		perror("kill");
		(void) sem_up(semid);
		return (5);
	}

	/* Wait for ACK/NAK signals */

	(void) sigsuspend(&mask);
	if (acknaksig == SIGNAK)
	{
		fprintf(stderr, "Client command failed\n");
		return (6);
	}

	/* Exit */

	return (0);
}
