/*
 *	Copyright 2011 Paul Parsons
 *
 *	This file is part of TinyInit.
 *
 *	TinyInit is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	TinyInit is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with TinyInit.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/wait.h>

#include "init.h"

/* Default halttype for when (1) service --stop=init (2) No more children */
static enum halttype ht_type = HT_REBOOT;

/*----------------------------------------------------------------------------*/

/*
 *	Service table.
 */

static struct servtab serv_tab[MAX_SERVICES];

static struct servtab *add_serv(const servname_t *sn, unsigned refcnt)
{
	struct servtab *s;

	for (s = &serv_tab[0]; s != &serv_tab[MAX_SERVICES]; ++s)
		if (s->state == SS_STOPPED)
			break;
	if (s == &serv_tab[MAX_SERVICES])
		return (NULL);

	s->sn = *sn;
	/* (s->state == SS_STOPPED) */
	s->refcnt = refcnt;
	s->ppid = -1;
	s->spid = -1;
	memset(&s->bmap, 0, sizeof(s->bmap));

	return (s);
}

static struct servtab *find_serv_by_name(const servname_t *sn)
{
	struct servtab *s;

	for (s = &serv_tab[0]; s != &serv_tab[MAX_SERVICES]; ++s)
		if (s->state != SS_STOPPED && memcmp(&s->sn, sn, sizeof(servname_t)) == 0)
			return (s);
	return (NULL);
}

static struct servtab *find_serv_by_ppid(pid_t ppid)
{
	struct servtab *s;

	for (s = &serv_tab[0]; s != &serv_tab[MAX_SERVICES]; ++s)
		if (s->state != SS_STOPPED && s->ppid == ppid)
			return (s);
	return (NULL);
}

static int for_each_child(struct servtab *s, int (*func)(struct servtab *s))
{
	int i;

	for (i = 0; i < BITMAPBANKS; ++i)
	{
		unsigned bank;

		bank = s->bmap[i];
		while (bank != 0)
		{
			int bit;
			struct servtab *r;

			bit = ffs(bank) - 1;
			r = &serv_tab[IDX_TO_BIT(i) + bit];
			if ((*func)(r) == -1)
				return (-1);
			bank &= ~(1 << bit);
		}
	}

	return (0);
}

/*----------------------------------------------------------------------------*/

/*
 *	Program execution.
 */

static char rlsingle[] = "RUNLEVEL=single";
static char rlnormal[] = "RUNLEVEL=normal";
static char *runlevel = rlnormal;
static char env_path[] = "PATH=/sbin:/usr/sbin:/bin:/usr/bin";
#if	defined(__gnu_linux__)
static char env_term[] = "TERM=linux";
#elif	defined(__FreeBSD__)
static char env_term[] = "TERM=cons25";
#elif	defined(__NetBSD__)
static char env_term[] = "TERM=vt100";
#elif	defined(__OpenBSD__)
static char env_term[] = "TERM=vt220";
#else
#error	"Unknown operating system"
#endif

static void exec_prog(char *name, char *arg1)
{
	sigset_t mask;
	char *argv[3];
	char *envp[4];

	(void) setsid();

	sigemptyset(&mask);
	(void) sigprocmask(SIG_SETMASK, &mask, NULL);

	argv[0] = name;
	argv[1] = arg1;
	argv[2] = NULL;

	envp[0] = runlevel;
	envp[1] = env_path;
	envp[2] = env_term;
	envp[3] = NULL;

	(void) execve(name, argv, envp);
	_exit(1);
	/* NOTREACHED */
}

static int run_prog(struct servtab *s, char *arg1, enum servstate state)
{
	pid_t pid;

	/* Cursory check for executable before fork() precludes error */
	if (access(s->sn.name, X_OK) == -1)
		return (-1);

	pid = fork();
	if (pid == -1)
		return (-1);
	if (pid == 0)
	{
		/* Child */
		exec_prog(s->sn.name, arg1);
		/* NOTREACHED */
	}
	/* Parent */
	s->state = state;
	s->ppid = pid;

	return (0);
}

static int start_prog(struct servtab *s)
{
	return (run_prog(s, "start", SS_STARTING));
}

static int stop_prog(struct servtab *s)
{
	return (run_prog(s, "stop", SS_STOPPING));
}

/*----------------------------------------------------------------------------*/

/*
 *	Service management: Client commands.
 */

static int add_children(struct servtab *s, const struct tinyshm *tiny)
{
	int i;

	for (i = 0; i < tiny->u.depend.count; ++i)
	{
		const servname_t *sn;
		struct servtab *r;
		int bit;
		unsigned *bankp;
		unsigned mask;

		sn = &tiny->u.depend.sn[i];

		r = find_serv_by_name(sn);
		if (! r)
		{
			r = add_serv(sn, 1);
			if (! r)
				return (-1);
			if (start_prog(r) == -1)
				return (-1);
		}
		else
			r->refcnt += 1;

		bit = (r - &serv_tab[0]);
		bankp = &s->bmap[BIT_TO_IDX(bit)];
		mask = (1 << (bit & (IDX_TO_BIT(1)-1)));
		if (! (*bankp & mask))
			*bankp |= mask;
		else	/* Duplicate */
			r->refcnt -= 1;
	}

	return (0);
}

static int can_resume(struct servtab *s)
{
	return ((s->state == SS_STOPPED || s->state == SS_STARTED) ? 0 : -1);
}

static int resume_ack(struct servtab *s)
{
	return ((s->state == SS_STARTED) ? 0 : -1);
}

static int tinydeps(const struct tinyshm *tiny)
{
	struct servtab *s;
	int sig;

	s = find_serv_by_ppid(tiny->u.depend.ppid);
	if (! s)	/* Process not mine */
		return (SIGNAK);

	switch (s->state)
	{
	case SS_STARTING:
		if (add_children(s, tiny) == -1)
			return (SIGNAK);

		if (for_each_child(s, can_resume) == -1)
		{
			s->state = SS_BLOCKED;
			s->spid = tiny->cpid;
			return (-1);
		}

		sig = (for_each_child(s, resume_ack) == 0) ? SIGACK : SIGNAK;
		break;
	case SS_BLOCKED:
		/* Prog error */
		sig = SIGNAK;
		break;
	case SS_STOPPING:
		/* Nothing to do */
		sig = SIGACK;
		break;
	default:
		/* NOTREACHED */
		sig = SIGNAK;
		break;
	}

	return (sig);
}

static int tinystart(const struct tinyshm *tiny)
{
	struct servtab *s;

	s = find_serv_by_name(&tiny->u.sn);
	if (s)
		return (SIGNAK);

	s = add_serv(&tiny->u.sn, 0);
	if (! s)
		return (SIGNAK);
	if (start_prog(s) == -1)
		return (SIGNAK);

	return (SIGACK);
}

static int tinystop(const struct tinyshm *tiny)
{
	struct servtab *s;

	s = find_serv_by_name(&tiny->u.sn);
	if (! s)
		return (SIGNAK);

	if (s->state != SS_STARTED || s->refcnt != 0)
		return (SIGNAK);

	if (stop_prog(s) == -1)
		return (SIGNAK);

	return (SIGACK);
}

static int tinylist(key_t key)
{
	int servid;
	void *serv;

	servid = shmget(key, sizeof(serv_tab), 0);
	if (servid == -1)
		return (SIGNAK);
	serv = shmat(servid, NULL, 0);
	if (serv == (void *)-1)
		return (SIGNAK);

	memcpy(serv, serv_tab, sizeof(serv_tab));

	(void) shmdt(serv);

	return (SIGACK);
}

static int tinyhalt(enum halttype type)
{
	struct servtab *s;

	ht_type = type;

	for (s = &serv_tab[0]; s != &serv_tab[MAX_SERVICES]; ++s)
		if (s->state == SS_STARTED && s->refcnt == 0)
			if (stop_prog(s) == -1)
				return (-1);

	return (0);
}

static int handle_ccmd(const struct tinyshm *tiny, int semid)
{
	int sig;

	switch (tiny->ccmd)
	{
	case CC_DEPEND:
		sig = tinydeps(tiny);
		break;
	case CC_START:
		sig = tinystart(tiny);
		break;
	case CC_STOP:
		sig = tinystop(tiny);
		break;
	case CC_LIST:
		sig = tinylist(tiny->u.key);
		break;
	case CC_HALT:
		sig = tinyhalt(tiny->u.type);
		break;
	default:
		sig = SIGNAK;
		break;
	}

	if (sig != -1)
		kill(tiny->cpid, sig);

	return (sem_up(semid));
}

/*----------------------------------------------------------------------------*/

/*
 *	Service management: SIGCHLD signals.
 */

static const enum servstate exit_tab[SS_count][2] =
{
		/* status==0, status!=0 */
[SS_STOPPED]=	{ SS_STOPPED, SS_STOPPED },	/* NOTREACHED */
[SS_STARTING]=	{ SS_STARTED, SS_STOPPED },
[SS_BLOCKED]=	{ SS_STARTED, SS_STOPPED },	/* Prog error */
[SS_STARTED]=	{ SS_STARTED, SS_STARTED },	/* NOTREACHED */
[SS_STOPPING]=	{ SS_STOPPED, SS_STARTED },
};

static int dec_refcnt(struct servtab *s)
{
	s->refcnt -= 1;
	return (0);
}

static void sigchld_boot(void)
{
	struct servtab *s;

	for (s = &serv_tab[0]; s != &serv_tab[MAX_SERVICES]; ++s)
	{
		int sig;

		if (s->state != SS_BLOCKED)
			continue;

		if (for_each_child(s, can_resume) == -1)
			continue;

		s->state = SS_STARTING;

		sig = (for_each_child(s, resume_ack) == 0) ? SIGACK : SIGNAK;
		(void) kill(s->spid, sig);
		s->spid = -1;	/* Beware of future pid wrap around */
	}
}

static int halt_child(struct servtab *s)
{
	if (s->state == SS_STARTED && s->refcnt == 0)
		if (stop_prog(s) == -1)
			return (-1);
	return (0);
}

static void sigchld_halt(struct servtab *s)
{
	if (for_each_child(s, halt_child) == -1)
		return;

	for (s = &serv_tab[0]; s != &serv_tab[MAX_SERVICES]; ++s)
		if (s->state != SS_STOPPED)
			return;

	myreboot(ht_type, 0);
	/* NOTREACHED */
}

static int handle_chld(void)
{
	int status;
	pid_t ppid;
	struct servtab *s;

again:
	ppid = waitpid(-1, &status, WNOHANG);
	if (ppid == -1 && errno == ECHILD)	/* No more children */
	{
		tinyhalt(HT_REBOOT);
		return (0);
	}
	if (ppid == -1)	/* Unexpected error */
		return (-1);
	if (ppid == 0)	/* No more zombies */
		return (0);

	s = find_serv_by_ppid(ppid);
	if (! s)	/* Process not mine */
		goto again;

	s->state = exit_tab[s->state][status != 0];

	if (s->state == SS_STOPPED)
		for_each_child(s, dec_refcnt);

	s->ppid = -1;	/* Beware of future pid wrap around */

	if (status == 0)
	{
		if (s->state == SS_STARTED)
			sigchld_boot();
		else
			sigchld_halt(s);
	}

	goto again;
}

/*----------------------------------------------------------------------------*/

/*
 *	Main.
 */

#if	defined(__gnu_linux__)
static const char singleuser[] = "S";
#elif	defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
static const char singleuser[] = "-s";
#else
#error	"Unknown operating system"
#endif

static const char devnull[] = "/dev/null";

static int pokesig;
static int chldsig;
static void handler(int sig)
{
	if (sig == SIGPOKE)
		pokesig = 1;
	else	/* (sig == SIGCHLD) */
		chldsig = 1;
}

static const char servdir[] = "/etc/init.d";

static const servname_t initname = { .name = "init" };

int main(int argc, char *argv[])
{
	int i;
	sigset_t mask;
	struct sigaction sa;
	int shmid;
	struct tinyshm *tiny;
	unsigned cseq;
	int semid;
	union { int val; struct semid_ds *buf; unsigned short *array; } arg;
	struct servtab *s;

	for (i = 1; i < argc; ++i)
		if (strcmp(argv[i], singleuser) == 0)
			runlevel = rlsingle;

	/* Create new session and disassociate any controlling terminal */
	if (setsid() == -1)
		return (1);

	/* Ensure child process opens do not alias to stdin/stdout/stderr */
	if (fcntl(STDIN_FILENO, F_GETFL) == -1)
		if (set_stdio(devnull) == -1)
			return (1);

	/* Block all signals */
	sigfillset(&mask);
	if (sigprocmask(SIG_SETMASK, &mask, NULL) == -1)
		return (2);

	/* Allow POKE/CHLD signals during sigsuspend() */
	sigdelset(&mask, SIGPOKE);
	sigdelset(&mask, SIGCHLD);

	/* Catch POKE/CHLD signals */
	sa.sa_handler = handler;
	sigemptyset(&sa.sa_mask);
	sigaddset(&sa.sa_mask, SIGPOKE);
	sigaddset(&sa.sa_mask, SIGCHLD);
	sa.sa_flags = SA_NOCLDSTOP;
	if (sigaction(SIGPOKE, &sa, NULL) == -1)
		return (2);
	if (sigaction(SIGCHLD, &sa, NULL) == -1)
		return (2);

	shmid = shmget(INIT_KEY, sizeof(struct tinyshm), (IPC_CREAT|IPC_EXCL|0600));
	if (shmid == -1)
		return (4);
	tiny = shmat(shmid, NULL, SHM_RDONLY);
	if (tiny == (struct tinyshm *)-1)
		return (4);
	cseq = tiny->cseq;	/* 0 */

	semid = semget(INIT_KEY, 1, (IPC_CREAT|IPC_EXCL|0600));
	if (semid == -1)
		return (3);
	arg.val = 1;
	if (semctl(semid, 0, SETVAL, arg) == -1)
		return (3);

	if (chdir(servdir) == -1)
		return (5);

	s = add_serv(&initname, 0);
	if (! s)
		return (6);
	if (start_prog(s) == -1)
		return (6);

	for ( ; ; )
	{
		/* Wait for POKE/CHLD signals */
		(void) sigsuspend(&mask);

		if (pokesig)
		{
			pokesig = 0;	/* reset flag */
			if (cseq != tiny->cseq)
			{
				cseq = tiny->cseq;
				if (handle_ccmd(tiny, semid) == -1)
					return (8);
			}
		}

		if (chldsig)
		{
			chldsig = 0;	/* reset flag */
			if (handle_chld() == -1)
				return (9);
		}
	}
}
