/*
 *	Copyright 2011 Paul Parsons
 *
 *	This file is part of TinyInit.
 *
 *	TinyInit is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	TinyInit is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with TinyInit.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <getopt.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <sys/sem.h>
#include <sys/shm.h>

#include "init.h"

static const struct option long_options[] =
{
	{	"depend",	0,	0,	'd'	},
	{	"start",	0,	0,	'S'	},
	{	"stop",		0,	0,	'K'	},
	{	"list",		0,	0,	'L'	},
	{	NULL,		0,	NULL,	0	}
};

#define	SERV_KEY	0x73657276	/* "serv" */

static int acknaksig;
static void handler(int sig)
{
	acknaksig = sig;
}

static const char *ss_tab[SS_count] =
{
[SS_STOPPED]=	"stopped",
[SS_STARTING]=	"starting",
[SS_BLOCKED]=	"blocked",
[SS_STARTED]=	"started",
[SS_STOPPING]=	"stopping",
};

static void pr_list(int servid)
{
	struct servtab *serv;
	int i;

	serv = shmat(servid, NULL, SHM_RDONLY);
	if (serv == (struct servtab *)-1)
	{
		perror("shmat");
		return;
	}

	printf("INDEX\tNAME\t\tSTATE\tREFCNT\tPPID\tSPID\tDEPEND\n");

	for (i = 0; i < MAX_SERVICES; ++i)
	{
		struct servtab *s;
		char c;
		int j;

		s = &serv[i];

		if (s->state == SS_STOPPED)
			continue;

		printf("%d\t%-16.32s%s\t%d\t%d\t%d",
			i,
			s->sn.name,
			ss_tab[s->state],
			s->refcnt,
			s->ppid,
			s->spid);

		c = '\t';
		for (j = 0; j < BITMAPBANKS; ++j)
		{
			unsigned bank;

			bank = s->bmap[j];
			while (bank != 0)
			{
				int bit;

				bit = ffs(bank) - 1;
				printf("%c%zu", c, (IDX_TO_BIT(j) + bit));
				c = ',';
				bank &= ~(1 << bit);
			}
		}

		printf("\n");
	}

	(void) shmdt(serv);
}

int main(int argc, char *argv[])
{
	enum clientcmd ccmd;
	int c;
	servname_t sn[MAX_DEPS];
	int i;
	sigset_t mask;
	struct sigaction sa;
	int shmid;
	struct tinyshm *tiny;
	int semid;
	int servid = -1;

	/* Process arguments */

	ccmd = CC_NULL;

	opterr = 0;
	while ((c = getopt_long(argc, argv, "+", long_options, NULL)) != -1)
	{
		switch (c)
		{
		case 'd':
			if (ccmd != CC_NULL)
				goto optdone;
			ccmd = CC_DEPEND;
			break;
		case 'S':
		case 'K':
			if (ccmd != CC_NULL)
				goto optdone;
			ccmd = (c == 'S') ? CC_START : CC_STOP;
			break;
		case 'L':
			if (ccmd != CC_NULL)
				goto optdone;
			ccmd = CC_LIST;
			break;
		}
		if (c == '?')
			break;
	}
optdone:

	if (c != -1
		|| ccmd == CC_NULL
		|| (ccmd == CC_DEPEND && optind == argc)
		|| (ccmd == CC_START && (argc - optind) != 1)
		|| (ccmd == CC_STOP && (argc - optind) != 1)
		|| (ccmd == CC_LIST && optind != argc))
	{
		fprintf(stderr, "usage:\tservice --depend service ...\n");
		fprintf(stderr, "\tservice --start service\n");
		fprintf(stderr, "\tservice --stop service\n");
		fprintf(stderr, "\tservice --list\n");
		return (1);
	}

	memset(sn, 0, sizeof(sn));
	i = 0;
	while (optind != argc)	/* CC_DEPEND, CC_START, CC_STOP */
	{
		char *s;

		if (i == MAX_DEPS)
		{
			fprintf(stderr, "Exceeded %d services\n", MAX_DEPS);
			return (1);
		}
		s = argv[optind];
		if (strlen(s) > (SERVNAMELEN-1))
		{
			fprintf(stderr, "Service name exceeds %d characters\n", (SERVNAMELEN-1));
			return (1);
		}
		strncpy(sn[i].name, s, SERVNAMELEN);
		++optind;
		++i;
	}

	/* Block ACK/NAK signals */

	sigemptyset(&mask);
	sigaddset(&mask, SIGACK);
	sigaddset(&mask, SIGNAK);
	if (sigprocmask(SIG_SETMASK, &mask, &mask) == -1)
		return (2);

	/* Catch ACK/NAK signals */

	sa.sa_handler = handler;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	if (sigaction(SIGACK, &sa, NULL) == -1)
		return (2);
	if (sigaction(SIGNAK, &sa, NULL) == -1)
		return (2);

	/* Get shm */

	shmid = shmget(INIT_KEY, sizeof(struct tinyshm), 0);
	if (shmid == -1)
	{
		perror("shmget");
		return (3);
	}
	tiny = shmat(shmid, NULL, 0);
	if (tiny == (struct tinyshm *)-1)
	{
		perror("shmat");
		return (3);
	}

	/* Get sem */

	semid = semget(INIT_KEY, 1, 0);
	if (semid == -1)
	{
		perror("semget");
		return (4);
	}
	if (sem_dn(semid) == -1)
	{
		perror("semop");
		return (4);
	}

	/* Write shm */

	tiny->ccmd = ccmd;
	tiny->cpid = getpid();
	if (ccmd == CC_DEPEND)
	{
		tiny->u.depend.ppid = getppid();
		tiny->u.depend.count = i;
		memcpy(tiny->u.depend.sn, sn, sizeof(sn));
	}
	else if (ccmd == CC_START || ccmd == CC_STOP)
	{
		tiny->u.sn = sn[0];
	}
	else	/* (ccmd == CC_LIST) */
	{
		servid = shmget(SERV_KEY,
			sizeof(struct servtab[MAX_SERVICES]),
			(IPC_CREAT|IPC_EXCL|0600));
		if (servid == -1)
		{
			perror("shmget");
			(void) sem_up(semid);
			return (3);
		}
		tiny->u.key = SERV_KEY;
	}
	tiny->cseq += 1;	/* Must be last field written */
	(void) shmdt(tiny);

	/* Poke init */

	if (kill(1, SIGPOKE) == -1)
	{
		perror("kill");
		if (ccmd == CC_LIST)
			(void) shmctl(servid, IPC_RMID, NULL);
		(void) sem_up(semid);
		return (5);
	}

	/* Wait for ACK/NAK signals */

	(void) sigsuspend(&mask);
	if (acknaksig == SIGNAK)
	{
		fprintf(stderr, "Client command failed\n");
		return (6);
	}

	/* Print list */

	if (ccmd == CC_LIST)
	{
		pr_list(servid);
		(void) shmctl(servid, IPC_RMID, NULL);
	}

	/* Exit */

	return (0);
}
