/*
 *	Copyright 2011 Paul Parsons
 *
 *	This file is part of TinyInit.
 *
 *	TinyInit is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	TinyInit is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with TinyInit.  If not, see <http://www.gnu.org/licenses/>.
 */

#if	defined(__gnu_linux__)
#define _XOPEN_SOURCE 700	/* Get crypt(3) declaration */
#endif

#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include <sys/ioctl.h>

#include "init.h"

static const struct option long_options[] =
{
	{	"port",		1,	0,	'p'	},
	{	NULL,		0,	NULL,	0	}
};

static int set_ctrlterm(void)
{
	int fd;

	fd = open("/dev/tty", O_RDONLY);
	if (fd == -1)
	{
		/* We do not have controlling terminal */
		if (ioctl(STDIN_FILENO, TIOCSCTTY, 0) == -1)
		{
			perror("TIOCSCTTY");
			return (-1);
		}
	}
	else
	{
		/* We do have controlling terminal */
		(void) close(fd);
	}

	return (0);
}

#if	defined(__gnu_linux__)
/* name:password:change:min:max:warning:inactivity:expire:reserved */
static const char etc_shadow[] = "/etc/shadow";
/* name:password:uid:gid:gecos:home:shell */
static const char etc_passwd[] = "/etc/passwd";
static const int field_pass = 2;
static const int field_home = 6;
static const int field_shell = 7;
#elif	defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
/* name:password:uid:gid:class:change:expire:gecos:home:shell */
static const char etc_shadow[] = "/etc/master.passwd";
/* name:password:uid:gid:gecos:home:shell */
static const char etc_passwd[] = "/etc/passwd";
static const int field_pass = 2;
static const int field_home = 6;
static const int field_shell = 7;
#else
#error	"Unknown operating system"
#endif

static char *lookup(const char *file, int field)
{
	static char staticbuf[1024];
	FILE *fp;
	char *p;

	fp = fopen(file, "r");
	if (! fp)
		return (NULL);

	while (fgets(staticbuf, sizeof(staticbuf), fp))
		if (strncmp(staticbuf, "root:", 5) == 0)
			break;

	(void) fclose(fp);

	if (strncmp(staticbuf, "root:", 5) != 0)
		return (NULL);

	p = strchr(staticbuf, '\n');
	if (p)
		*p = '\0';

	p = staticbuf;

	for ( ; ; )
	{
		char *s;

		s = strchr(p, ':');
		if (s)
			*s++ = '\0';
		if (field == 1)
			return (p);
		if (! s)
			return (NULL);
		--field;
		p = s;
	}
}

static int set_echo(tcflag_t echo)
{
	struct termios termios;

	if (tcgetattr(STDIN_FILENO, &termios) == -1)
	{
		perror("tcgetattr");
		return (-1);
	}
	termios.c_lflag = (termios.c_lflag & ~ECHO) | echo;
	if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &termios) == -1)
	{
		perror("tcsetattr");
		return (-1);
	}
	return (0);
}

static int getpasswd(char *passwd, size_t size)
{
	for ( ; ; )
	{
		ssize_t n;
		char *s;

		printf("root password: ");
		fflush(stdout);

		n = read(STDIN_FILENO, passwd, size);

		printf("\n");
		fflush(stdout);

		if (n == -1)
		{
			perror("read");
			return (-1);
		}
		if (n == 0)	/* EOF */
			continue;
		if (n == size)	/* Too much */
			continue;

		passwd[n] = '\0';

		s = strchr(passwd, '\n');
		if (s)
			*s = '\0';

		if (s != passwd)
			return (0);
	}
}

extern char **environ;

static char *search_env(const char *name)
{
	size_t len;
	char **envp;
	char *s;

	len = strlen(name);
	envp = environ;
	while ((s = *envp++) != NULL)
		if (strncmp(s, name, len) == 0 && s[len] == '=')
			break;
	return (s);
}

int main(int argc, char *argv[])
{
	const char *port;
	int c;
	char *s;
	char env_home[128];
	char shell[64];
	char *shargv[3];
	char *shenvp[4];

	port = NULL;

	opterr = 0;
	while ((c = getopt_long(argc, argv, "+", long_options, NULL)) != -1)
	{
		switch (c)
		{
		case 'p':
			port = optarg;
			if (strlen(port) == 0)	/* i.e. "--port=" */
				goto optdone;
			break;
		}
		if (c == '?')
			break;
	}
optdone:

	if (c != -1 || optind != argc)
	{
		fprintf(stderr, "usage:\tsulogin [--port=port]\n");
		return (1);
	}

	if (port)
	{
		char dev[64];

		if (strlen(port) > (sizeof(dev)-5-1))
		{
			fprintf(stderr, "Port path exceeds %zu characters\n", (sizeof(dev)-1));
			return (1);
		}
		strcpy(dev, "/dev/");
		strcat(dev, port);

		if (set_stdio(dev) == -1)
			return (1);

		if (set_ctrlterm() == -1)
			return (1);
	}

	s = lookup(etc_shadow, field_pass);
	if (! s)
		s = lookup(etc_passwd, field_pass);
	if (s && strlen(s) != 0)
	{
		char passwd[64];
		int ret;

again:
		if (set_echo(0) == -1)
			return (1);
		ret = getpasswd(passwd, sizeof(passwd));
		if (set_echo(ECHO) == -1)
			return (1);
		if (ret == -1)
			return (1);
		if (strcmp(crypt(passwd, s), s) != 0)
		{
			printf("Login incorrect\n");
			goto again;
		}
	}

	strcpy(env_home, "HOME=");
	s = lookup(etc_passwd, field_home);
	if (! s || strlen(s) == 0 || strlen(s) > (sizeof(env_home)-5-1))
		strcat(env_home, "/");
	else
		strcat(env_home, s);

	s = lookup(etc_passwd, field_shell);
	if (! s || strlen(s) == 0 || strlen(s) > (sizeof(shell)-1))
		strcpy(shell, "/bin/sh");
	else
		strcpy(shell, s);

	shargv[0] = shell;
	shargv[1] = "-i";	/* Interactive */
	shargv[2] = NULL;

	shenvp[0] = env_home;
	shenvp[1] = search_env("PATH");
	shenvp[2] = search_env("TERM");	/* May be NULL */
	shenvp[3] = NULL;

	(void) execve(shell, shargv, shenvp);
	perror("execve");
	return (1);
}
