/*
 *	Copyright 2011 Paul Parsons
 *
 *	This file is part of TinyInit.
 *
 *	TinyInit is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	TinyInit is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with TinyInit.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <sys/wait.h>

/*
 *	Test whether sigsuspend() handles SIGCHLD.
 *
 *	Results from selected kernels:
 *
 *	Linux 2.6.39:	PASS
 *
 *	FreeBSD 8.2:	PASS
 *
 *	NetBSD 5.1:	PASS
 *
 *	OpenBSD 4.9:	PASS
 */

static int chldsig;
static void handler(int sig)
{
	chldsig = 1;
}

static int test(void)
{
	sigset_t mask;
	struct sigaction sa;
	pid_t pid;
	int status;

	/* Block CHLD signals */
	sigemptyset(&mask);
	sigaddset(&mask, SIGCHLD);
	if (sigprocmask(SIG_SETMASK, &mask, &mask) == -1)
	{
		perror("sigprocmask");
		return (-1);
	}

	/* Catch CHLD signals */
	sa.sa_handler = handler;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_NOCLDSTOP;
	if (sigaction(SIGCHLD, &sa, NULL) == -1)
	{
		perror("sigaction");
		return (-1);
	}

	pid = fork();
	switch (pid)
	{
	case -1:
		perror("fork");
		return (-1);
	case 0:
		/* Child */
		(void) sleep(2);
		_exit(0);
	default:
		/* Parent */
		printf("fork: pid=%d\n", pid);
		break;
	}

	/* Wait for CHLD signals */
	(void) sigsuspend(&mask);

	if (chldsig == 0)
	{
		printf("Spurious signal\n");
		return (-1);
	}

	pid = waitpid(pid, &status, 0);
	if (pid == -1)
	{
		printf("waitpid: %s\n", strerror(errno));
		return (-1);
	}

	printf("pid=%d, status=0x%04x\n", pid, status);
	return (0);
}

int main(void)
{
	int ret;

	ret = test();
	printf("TEST: %s\n", (ret == -1) ? "FAIL" : "PASS");

	return (0);
}
