/*
 *	Copyright 2011 Paul Parsons
 *
 *	This file is part of TinyInit.
 *
 *	TinyInit is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	TinyInit is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with TinyInit.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <sys/wait.h>

/*
 *	Test whether sigwait() handles SIGCHLD.
 *
 *	Results from selected kernels:
 *
 *	Linux 2.6.39:
 *		SIG_IGN FAIL
 *		SIG_DFL PASS
 *		handler PASS
 *
 *	FreeBSD 8.2:
 *		SIG_IGN FAIL
 *		SIG_DFL FAIL
 *		handler PASS
 *
 *	NetBSD 5.1:
 *		SIG_IGN FAIL
 *		SIG_DFL FAIL
 *		handler PASS
 *
 *	OpenBSD 4.9:
 *		SIG_IGN FAIL
 *		SIG_DFL FAIL
 *		handler FAIL
 */

static void sighandler(int sig)
{
	;	/* Nothing to do */
}

static int test(void (*handler)(int))
{
	sigset_t mask;
	struct sigaction sa;
	pid_t pid;
	int ret;
	int sig;
	int status;

	/* Block ALRM/CHLD signals */
	sigemptyset(&mask);
	sigaddset(&mask, SIGALRM);
	sigaddset(&mask, SIGCHLD);
	if (sigprocmask(SIG_SETMASK, &mask, NULL) == -1)
	{
		perror("sigprocmask");
		return (-1);
	}

	/* Catch CHLD signals */
	sa.sa_handler = handler;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_NOCLDSTOP;
	if (sigaction(SIGCHLD, &sa, NULL) == -1)
	{
		perror("sigaction");
		return (-1);
	}

	pid = fork();
	switch (pid)
	{
	case -1:
		perror("fork");
		return (-1);
	case 0:
		/* Child */
		(void) sleep(2);
		_exit(0);
	default:
		/* Parent */
		printf("fork: pid=%d\n", pid);
		(void) alarm(4);
		break;
	}

	/* Wait for ALRM/CHLD signals */
	ret = sigwait(&mask, &sig);
	if (ret != 0)
	{
		printf("sigwait: %s\n", strerror(ret));
		return (-1);
	}

	if (sig == SIGALRM)
	{
		printf("Alarm: kill(pid=%d, sig=%d)\n", pid, SIGTERM);
		if (kill(pid, SIGTERM) == -1)
			printf("Alarm: kill: %s\n", strerror(errno));
	}

	pid = waitpid(pid, &status, 0);
	if (pid == -1)
	{
		printf("waitpid: %s\n", strerror(errno));
		return (-1);
	}

	switch (sig)
	{
	case SIGALRM:
		printf("Alarm: pid=%d, status=0x%04x\n", pid, status);
		return (-1);
	case SIGCHLD:
		printf("OK\n");
		return (0);
	default:
		printf("sigwait() returned sig=%d\n", sig);
		return (-1);
	}
}

int main(void)
{
	int ret;

	printf("TEST: Testing SIG_IGN...\n");
	ret = test(SIG_IGN);
	printf("TEST: %s\n", (ret == -1) ? "FAIL" : "PASS");

	printf("TEST: Testing SIG_DFL...\n");
	ret = test(SIG_DFL);
	printf("TEST: %s\n", (ret == -1) ? "FAIL" : "PASS");

	printf("TEST: Testing handler...\n");
	ret = test(sighandler);
	printf("TEST: %s\n", (ret == -1) ? "FAIL" : "PASS");

	return (0);
}
